# dzn-runtime -- Dezyne runtime library
#
# Copyright © 2016, 2017, 2019, 2020, 2021, 2022, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
# Copyright © 2017 Jvaneerd <J.vaneerd@student.fontys.nl>
# Copyright © 2017, 2018, 2019, 2021, 2022 Rutger van Beusekom <rutger@dezyne.org>
# Copyright © 2016 Henk Katerberg <hank@mudball.nl>
# Copyright © 2024 Paul Hoogendijk <paul@dezyne.org>
#
# This file is part of dzn-runtime.
#
# All rights reserved.
#
#
# Commentary:
#
# Code:

import sys


def bool_to_string(bool):
    return str(bool).lower()

def int_to_string(int):
    return str(int)

def enum_to_string(enum_to_string, value):
    return enum_to_string[value]

##### TODO move to generated main!!!

def event_value_str(event):
    return event.split(".")[-1]

def event_to_bool(event):
    to_bool = {"false": False, "true": True}
    return to_bool[event_value_str(event)]

def event_to_int(event):
    return int(event_value_str(event))

def event_enum_field_str(event):
    return event.split(":")[-1]

def event_to_enum(type, event):
    return type.__dict__.get(event_enum_field_str(event), None)

######


class V:
    def __init__ (self, v):
        self.v = v

def illegal ():
    raise RuntimeError ('illegal')

class Runtime:
    def __init__ (self, illegal=illegal):
        self.components = []
        self.illegal = illegal
        self.info = {}
    def flushes (self, c, enable=True):
        self.components += [c]
        self.info[c] = self.info.get (c, Info ())
        self.info[c].flushes = enable
    def external_p (self, c):
        return c not in self.components
    def flush (self, c):
        if (self.external_p (c)):
            return
        while (self.info[c].q):
            self.handle (c, self.info[c].q.pop ())
        if (self.info[c].deferred):
            t = self.info[c].deferred
            self.info[c].deferred = None;
            if (not self.info[t].handling):
                self.flush (t)
    def defer (self, i, o, f):
        if (not (i and self.info[i].flushes) and not self.info[o].handling):
            self.handle (o, f)
        else:
            self.info[o].q.insert (0, f)
            if i:
                self.info[i].deferred = o
    def valued_helper (self, c, f, m):
        if (self.info[c].handling):
            raise Exception ('a valued event cannot be deferred')
        self.info[c].handling = True
        r = f ()
        self.info[c].handling = False
        self.flush (c)
        return r
    def handle (self, c, f):
        if (not self.info[c].handling):
            self.info[c].handling = True
            f ()
            self.info[c].handling = False
            self.flush (c)
        else:
            raise Exception ('component already handling an event')

class Port:
    def __init__ (self, name='', component=None):
        self.dzn_name = name
        self.dzn_component = component

class Component:
    def __init__ (self, locator, name='', parent=None):
        self.dzn_locator = locator
        self.dzn_runtime = locator.get (Runtime)
        self.dzn_name = name
        self.dzn_parent = parent

class Info:
    def __init__ (self):
        self.handling = False
        self.deferred = None
        self.flushes = False
        self.q = []

def call_in (c, f, m):
    trace_in (m[0], m[1])
    c.dzn_runtime.handle (c, f)
    trace_out (m[0], 'return')

def rcall_in (c, f, m):
    trace_in (m[0], m[1])
    r = c.dzn_runtime.valued_helper (c, f, m)
    trace_out (m[0], m[2](r))
    return r

def call_out (c, f, m):
    trace_out (m[0], m[1])
    c.dzn_runtime.defer (m[0].in_port.dzn_component, c, f)

def path (m, p=''):
    if (not m):
        return '<external>' + ('.' + p if p else p);
    if ('dzn_component' in m.__dict__.keys ()):
        return path (m.dzn_component, m.dzn_name + ('.' + p if p else p))
    if ('dzn_parent' in m.__dict__.keys () and m.dzn_parent):
        return path (m.dzn_parent, m.dzn_name + ('.' + p if p else p))
    return m.dzn_name + ('.' + p if p and m.dzn_name else p)

def trace_in (i, e):
    sys.stderr.write (path (i.out_port) + '.' + e + ' -> '
                      + path (i.in_port) + '.' + e + '\n')

def trace_out (i, e):
    sys.stderr.write (path(i.out_port) + '.' + e + ' <- '
                      + path(i.in_port) + '.' + e + '\n')

def connect (provided, required):
    provided.out_port = required.out_port
    required.in_port = provided.in_port

def mixin_meta (port, port_meta):
    new_port = port
    new_port.in_port.dzn_name = port_meta.in_port.dzn_name
    new_port.out_port.dzn_name = port_meta.out_port.dzn_name
    new_port.in_port.dzn_component = port_meta.in_port.dzn_component
    new_port.out_port.dzn_component = port_meta.out_port.dzn_component
    return new_port
# version: 2.18.2.230-5e322
